import time
from enum import IntEnum
from typing import Callable, Generic, List, Tuple, Type, TypeVar, Union

from marshmallow import Schema  # type: ignore

from pysjtu.utils import overlap, parse_slice, range_in_set


class Gender(IntEnum):
    male = 1
    female = 2


class Result:
    """ Base class for Result """

    def __repr__(self):
        raise NotImplementedError  # pragma: no cover


class PARTIAL:
    pass


class LazyResult(Result):
    """ Base class for LazyResult """
    _load_func: Callable = None

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        if value == PARTIAL:
            update_dict = self._load_func()
            for k, v in update_dict.items():
                super().__setattr__(k, v)
            value = super().__getattribute__(item)
        return value


T_Result = TypeVar("T_Result", bound=Result)


class QueryResult(Generic[T_Result]):
    """
    A key accessible, sliceable, and iterable interface to query result collections.

    A QueryResult object is constructed with a raw data callable reference.

    A QueryResult object is returned by a query operation, and isn't meant to be constructed by a user.

    A QueryResult object is lazy, which means network I/Os won't be performed until items are actually accessed.

    :param method_ref: The request method to be called when fetching data.
    :param post_ref: The schema load method to be called on fetched data.
    :param query_params: Parameters for this query.
    :param page_size: The page size for result iteration.
    """
    _ref: Callable
    _post_ref: Callable
    _query_params: dict
    _length: int
    _cache: List[dict]
    _cached_items: set
    _page_size: int

    def __init__(self, method_ref: Callable, post_ref: Callable, query_params: dict, page_size: int = 15):
        self._ref = method_ref  # type: ignore
        self._post_ref = post_ref  # type: ignore
        self._query_params = query_params
        self._length = 0
        # noinspection PyTypeChecker
        self._cache = [{}] * len(self)
        self._cached_items = set()
        self._page_size = page_size

    def __getitem__(self, arg: Union[int, slice]) -> T_Result:
        if isinstance(arg, int):
            data = self._handle_result_by_index(arg)  # type: ignore
        elif isinstance(arg, slice):
            data = self._handle_result_by_idx_slice(arg)  # type: ignore
        else:
            raise TypeError("QueryResult indices must be integers or slices, not " + type(arg).__name__)
        data = self._post_ref(data)  # type: ignore
        return data  # type: ignore

    def _handle_result_by_index(self, idx: int) -> dict:
        idx = len(self) + idx if idx < 0 else idx
        if idx >= len(self) or idx < 0:
            raise IndexError("index out of range")
        self._update_cache(idx, idx + 1)
        return self._cache[idx]

    def _handle_result_by_idx_slice(self, idx: slice) -> list:
        idx_start = parse_slice(idx.start)
        idx_stop = parse_slice(idx.stop)

        if idx_start is None:
            start = 0
        elif idx_start < 0:
            start = len(self) + idx.start
        else:
            start = idx.start

        if idx_stop is None:
            end = len(self) - 1
        elif idx_stop < 0:
            end = len(self) + idx.stop - 1
        else:
            end = idx.stop

        if end > len(self):
            end = len(self)
        if start >= end:
            return []
        self._update_cache(start, end)
        return self._cache[idx]

    def __len__(self) -> int:
        if not self._length:
            rtn = self._query(1, 1)
            self._length = rtn["totalResult"]
        return self._length

    def flush_cache(self):
        """ Flush caches. Local caches are dropped and data will be fetched from remote. """
        self._length = 0
        # noinspection PyTypeChecker
        self._cache = [None] * len(self)
        self._cached_items = set()

    def _update_cache(self, start: int, end: int):
        fetch_set = set(set(range(start, end)) - self._cached_items)
        while len(fetch_set) != 0:
            fetch_range = next(range_in_set(fetch_set))
            page = int(fetch_range.start / self._page_size) + 1
            self._cached_items.update(range(*self._fetch_range(page, self._page_size)))
            fetch_set = set(set(range(start, end)) - self._cached_items)

    def _fetch_range(self, page: int, count: int) -> Tuple[int, int]:
        rtn = self._query(page, count)["items"]
        for item in zip(range(count * (page - 1), count * (page - 1) + len(rtn)), rtn):
            self._cache[item[0]] = item[1]
        return count * (page - 1), count * (page - 1) + len(rtn)

    def _query(self, page: int, count: int) -> dict:
        new_params = self._query_params
        new_params["queryModel.showCount"] = count
        new_params["queryModel.currentPage"] = page
        new_params["queryModel.sortName"] = ""
        new_params["queryModel.sortOrder"] = "asc"
        new_params["nd"] = int(time.time() * 1000)
        new_params["_search"] = False
        return self._ref(data=new_params).json()

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]


class Results(List[T_Result]):
    """
    Base class for Results

    :param year: year of the :class:`Results` object.
    :param term: term of the :class:`Results` object.
    """
    _schema: Type[Schema]
    _result_model: Type[T_Result]
    _valid_fields: List[str]

    def __init__(self, year: int = 0, term: int = 0):
        super().__init__()
        self._year = year
        self._term = term
        self._valid_fields = list(self._result_model.__annotations__.keys())

    @property
    def year(self) -> int:
        return self._year

    @property
    def term(self) -> int:
        return self._term

    def load(self, data: dict):
        """
        Load a list of dicts into Results, and deserialize dicts to Result objects.

        :param data: a list of dicts.
        """
        schema = self._schema(many=True)
        results = schema.load(data)
        for result in results:
            self.append(result)

    def filter(self, **param) -> List[T_Result]:
        """
        Get Result objects matching specific criteria.

        :param param: query criteria
        :return: Result objects matching given criteria.
        """
        rtn = list(self)
        for (k, v) in param.items():
            if k not in self._valid_fields:
                raise KeyError("Invalid criteria!")
            if k in ("week", "time", "day"):
                rtn = list(filter(lambda x: overlap(getattr(x, k), v), rtn))
            else:
                rtn = list(filter(lambda x: getattr(x, k) == v, rtn))
        return rtn



import typing
from copy import deepcopy

from marshmallow import EXCLUDE, Schema, fields, post_dump, post_load, pre_dump, pre_load  # type: ignore

from pysjtu.schemas.base import SplitField
from pysjtu.utils import replace_keys


class HasRoll(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        if "1" in value.split(","):
            return True
        if "2" in value.split(","):  # pragma: no cover
            return False  # pragma: no cover
        return None  # pragma: no cover

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if value is True:
            return 1
        if value is False:
            return 0
        return -1  # pragma: no cover


class Registered(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        if "3" in value.split(","):
            return True  # pragma: no cover
        if "4" in value.split(","):
            return False  # pragma: no cover
        return None

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if value is True:
            return 1
        if value is False:  # pragma: no cover
            return 0  # pragma: no cover
        return -1  # pragma: no cover


class Attending(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        if "5" in value.split(","):
            return True
        if "6" in value.split(","):  # pragma: no cover
            return False  # pragma: no cover
        return None  # pragma: no cover

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if value is True:
            return 1
        if value is False:
            return 0  # pragma: no cover
        return -1


class ConditionLogic(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if value == 0:  # pragma: no cover
            return LogicEnum.AND  # pragma: no cover
        if value == 1:  # pragma: no cover
            return LogicEnum.OR  # pragma: no cover
        return None  # pragma: no cover

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if not isinstance(value, LogicEnum):
            raise TypeError
        if value == LogicEnum.AND:
            return 0
        if value == LogicEnum.OR:  # pragma: no cover
            return 1  # pragma: no cover


class MakeupAsPass(fields.Field):
    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        return "bk" if value else ""


class RebuildAsPass(fields.Field):
    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        return "cx" if value else ""


class RankingField(fields.Field):
    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if not isinstance(value, Ranking):
            raise TypeError
        return value.value


class CourseRangeField(fields.Field):
    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if not isinstance(value, CourseRange):
            raise TypeError
        return value.value


class Percentage(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return float(value.replace("%", "")) / 100


class RankingResultField(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return int(value.split("/")[0])


class StudentCountFromRanking(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return int(value.split("/")[1])


class GPAQueryParamsSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    start_term = fields.Int(required=True, dump_key="qsXnxq", dump_only=True)
    end_term = fields.Int(required=True, dump_key="zzXnxq", dump_only=True)
    condition_logic = ConditionLogic(required=True, dump_key="tjgx", dump_only=True)
    makeup_as_60 = MakeupAsPass(required=True, dump_only=True)
    rebuild_as_60 = RebuildAsPass(required=True, dump_only=True)
    gp_round = fields.Int(required=True, load_key="cjblws", dump_key="sspjfblws")
    gpa_round = fields.Int(required=True, load_key="jdblws", dump_key="pjjdblws")
    exclude_gp = fields.Str(required=True, data_key="bjjd")
    exclude_gpa = fields.Str(required=True, data_key="bjpjf")
    course_whole = SplitField(required=True, load_key="tjqckc", dump_key="kch_ids", sep=",")
    course_range = CourseRangeField(required=True, dump_key="kcfw", dump_only=True)
    ranking = RankingField(required=True, dump_key="tjfw", dump_only=True)
    has_roll = HasRoll(required=True, data_key="atjc", load_only=True)
    registered = Registered(required=True, data_key="atjc", load_only=True)
    attending = Attending(required=True, data_key="atjc", load_only=True)
    has_roll_dump = HasRoll(required=True, data_key="xjzt", dump_only=True)
    registered_dump = Registered(required=True, data_key="zczt", dump_only=True)
    attending_dump = Attending(required=True, data_key="sfzx", dump_only=True)

    # noinspection PyUnusedLocal
    @pre_load
    def wrap_pre_load(self, data, **kwargs):
        data = {item["zdm"]: item["szz"] for item in filter(lambda x: "szz" in x.keys(), data)}
        pairs = tuple(
            (field.metadata['load_key'], field.data_key or field_name)
            for field_name, field in self.fields.items() if 'load_key' in field.metadata
        )
        return replace_keys(data, pairs)

    # noinspection PyUnusedLocal
    @post_load
    def wrap_post_load(self, data, **kwargs):
        return GPAQueryParams(makeup_as_60=False, rebuild_as_60=False, ranking=Ranking.GRADE_AND_FIELD,
                              course_range=CourseRange.CORE, start_term=None, end_term=None,
                              condition_logic=LogicEnum.AND, **data)

    # noinspection PyUnusedLocal
    @pre_dump
    def wrap_pre_dump(self, data, **kwargs):
        pre_dict = deepcopy(data.__dict__)
        pre_dict["has_roll_dump"] = pre_dict.pop("has_roll")
        pre_dict["registered_dump"] = pre_dict.pop("registered")
        pre_dict["attending_dump"] = pre_dict.pop("attending")
        return pre_dict

    # noinspection PyUnusedLocal
    @post_dump
    def wrap_post_dump(self, data, **kwargs):
        pairs = tuple(
            (field.data_key or field_name, field.metadata['dump_key'])
            for field_name, field in self.fields.items() if 'dump_key' in field.metadata
        )
        data = replace_keys(data, pairs)

        parse_fields = ["xjzt", "zczt", "sfzx"]
        data["alsfj"] = data.pop("makeup_as_60") + data.pop("rebuild_as_60")
        for field in parse_fields:
            if data[field] == -1:
                data.pop(field)

        parse_fields = ["qsXnxq", "zzXnxq"]
        for field in parse_fields:
            if data[field] is None:
                data[field] = ''
        return data


class GPASchema(Schema):
    class Meta:
        unknown = EXCLUDE

    total_score = fields.Int(required=True, data_key="zf")
    course_count = fields.Int(required=True, data_key="ms")
    fail_count = fields.Int(required=True, data_key="bjgms")
    total_credit = fields.Float(required=True, data_key="zxf")
    acquired_credit = fields.Float(required=True, data_key="hdxf")
    failed_credit = fields.Float(required=True, data_key="bjgxf")
    pass_rate = Percentage(required=True, data_key="tgl")
    gp = fields.Float(required=True, data_key="xjf")
    gp_ranking = RankingResultField(required=True, data_key="xjfpm")
    gpa = fields.Float(required=True, data_key="gpa")
    gpa_ranking = RankingResultField(required=True, data_key="gpapm", load_only=True)
    total_students = StudentCountFromRanking(required=True, data_key="gpapm", load_only=True)

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return GPA(**data)


from pysjtu.models.gpa import LogicEnum, Ranking, CourseRange, GPAQueryParams, GPA



from os import path

from flask import Flask, redirect, request, session

STATIC_DIR = path.join(path.dirname(path.abspath(__file__)), 'resources/website')

app = Flask(__name__, static_folder=STATIC_DIR)
app.secret_key = b'j(92IdXM+z_3xec]G'

UUID = "ad7283af-2198-476f-95ac-fa9356a0357c"
SU = "519027910001"


@app.route("/jaccount/captcha")
def login_captcha():
    return "OK"


@app.route("/jaccount/jalogin")
def login_ja():
    return "OK"


@app.route("/xtgl/login_slogin.html")
def login_isjtu():
    return app.send_static_file("login_slogin.html")


@app.route("/jaccountlogin")
def login():
    return app.send_static_file("jalogin")


@app.route("/login_patch")
def login_patch():
    session["is_login"] = True
    return redirect("/xtgl/index_initMenu.html")


@app.route("/jaccount/ulogin", methods=["post"])
def login_post():
    v = request.args.get("v")
    uuid = request.args.get("uuid")
    username = request.args.get("user")
    password = request.args.get("pass")
    captcha = request.args.get("captcha")
    if uuid == UUID and v == "" and username == "FeiLin" and password == "WHISPERS" and captcha == "ipsum":
        return redirect("https://i.sjtu.edu.cn/login_patch")
    if password == "1145141919810":
        return redirect("/jaccount/jalogin?err=0")
    return redirect("/jaccount/jalogin?err=1")


@app.route("/expire_me")
def session_expire():
    session["is_login"] = False
    return "OK"


@app.route("/is_login")
def is_login():
    return "True" if session.get("is_login", None) else "False"


@app.route("/logout")
def logout():
    session["is_login"] = False
    return redirect("/xtgl/login_slogin.html")


@app.route("/503")
def service_unavailable():
    return "ServiceUnavailable", 503


@app.route("/404")
def not_found():
    return "NotFound", 404


@app.route("/ping", methods=["GET", "HEAD", "PUT", "PATCH", "DELETE"])
def get_and_head():
    return "pong"


@app.route("/ping", methods=["POST"])
def post():
    return request.data


# --------------------------------------
@app.route("/xtgl/index_initMenu.html")
def homepage():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    return app.send_static_file("index_initMenu.html")


@app.route("/xtgl/index_cxshjdAreaFive.html")
def calendar():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    return app.send_static_file("index_cxshjdAreaFive.html")


@app.route("/cjgl/common_cxGnzdxxList.html")
def gpa_query_params():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.args.get("su", None) != SU:
        return ""
    return app.send_static_file("common_cxGnzdxxList.html")


@app.route("/kbcx/xskbcx_cxXsKb.html", methods=["post"])
def schedule():
    _args = ["xqm", "xnm"]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    return app.send_static_file("xskbcx_cxXsKb.html")


@app.route("/cjcx/cjcx_cxXsXmcjList.html", methods=["post"])
def score_detail():
    _args = ["xnm", "xqm", "jxb_id", "_search", "nd", "queryModel.showCount", "queryModel.currentPage",
             "queryModel.sortName", "queryModel.sortOrder", "time"]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    return app.send_static_file("cjcx_cxXsXmcjList.html")


@app.route("/cjcx/cjcx_cxDgXscj.html", methods=["post"])
def score():
    _args = ["xnm", "xqm", "_search", "nd", "queryModel.showCount", "queryModel.currentPage",
             "queryModel.sortName", "queryModel.sortOrder", "time"]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    return app.send_static_file("cjcx_cxDgXscj.html")


@app.route("/kwgl/kscx_cxXsksxxIndex.html", methods=["post"])
def exam():
    _args = ["xnm", "xqm", "ksmcdmb_id", "kch", "kc", "ksrq", "kkbm_id", "_search", "nd", "queryModel.showCount",
             "queryModel.currentPage", "queryModel.sortName", "queryModel.sortOrder", "time"]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    return app.send_static_file("kscx_cxXsksxxIndex.html")


@app.route("/design/funcData_cxFuncDataList.html", methods=["post"])
def course_lib():
    _args = ["xnm", "xqm", "_search", "nd", "queryModel.showCount", "queryModel.currentPage", "queryModel.sortName",
             "queryModel.sortOrder"]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    if request.args.get("func_widget_guid", None) != "DA1B5BB30E1F4CB99D1F6F526537777B":
        return ""
    page = request.form.get("queryModel.currentPage")
    count = request.form.get("queryModel.showCount")
    if count == '1':
        return app.send_static_file("funcData_cxFuncDataList_initial.html")
    if count == '40':
        return app.send_static_file(f"funcData_cxFuncDataList_{page}.html")
    return ""


@app.route("/cjpmtj/gpapmtj_tjGpapmtj.html", methods=["post"])
def gpa_query():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.form.get("tjgx", None) != "1":
        return app.send_static_file("gpapmtj_tjGpapmtj_unauthorized.html")
    if request.form.get("kcfw", None) != "hxkc":
        return app.send_static_file("gpapmtj_tjGpapmtj_failure.html")
    session["is_gpa_calculated"] = True
    return app.send_static_file("gpapmtj_tjGpapmtj.html")


@app.route("/cjpmtj/gpapmtj_cxGpaxjfcxIndex.html", methods=["post"])
def gpa():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if not session.get("is_gpa_calculated", None):
        return ""
    return app.send_static_file("gpapmtj_cxGpaxjfcxIndex.html")


@app.route("/test_selection")
def enable_selection():
    session["enable_selection"] = True
    return "OK"


@app.route("/test_no_conflict")
def no_conflict():
    session["no_conflict"] = True
    return "OK"


@app.route("/test_no_full")
def no_full():
    session["no_full"] = True
    return "OK"


@app.route("/get_session")
def get_session():
    return str(session.get(request.args.get("key", ""), ""))


@app.route("/xsxk/zzxkyzb_cxZzxkYzbIndex.html", methods=["get"])
def selection_all_sectors():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.args.get("su", None) != SU:
        return ""
    if not session.get("enable_selection", None):
        return app.send_static_file("zzxkyzb_cxZzxkYzbIndex_not_selecting.html")
    session["query_all_sectors"] = session.get("query_all_sectors", 0) + 1
    return app.send_static_file("zzxkyzb_cxZzxkYzbIndex.html")


@app.route("/xsxk/zzxkyzb_cxZzxkYzbDisplay.html", methods=["post"])
def selection_sector_param():
    _args = ["xkkz_id", "xszxzt", "kspage", "jspage"]
    sectors = [
        "A000000000000B76E055F8163ED16360",
        "A000000000000A7FE055F8163ED16360",
        "A000000000000881E055F8163ED16360",
        "A000000000000F12E055F8163ED16360",
        "A000000000000455E055F8163ED16360",
        "A000000000000B28E055F8163ED16360"
    ]
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.args.get("su", None) != SU:
        return ""
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    if request.form["xkkz_id"] not in sectors or request.form["xszxzt"] != "1":
        return ""
    session["query_sector_param"] = session.get("query_sector_param", 0) + 1
    return app.send_static_file("zzxkyzb_cxZzxkYzbDisplay.html")


@app.route("/xsxk/zzxkyzb_cxZzxkYzbPartDisplay.html", methods=["post"])
def selection_query_courses():
    _args = ['rwlx', 'xkly', 'xqh_id', 'zyh_id', 'njdm_id', 'bh_id', 'tykczgxdcs', 'xkxnm', 'xkxqm', 'kklxdm', 'xslbdm',
             'ccdm', 'kklxdm', 'xbm', 'zyfx_id', 'xsbj', 'kkbk', 'sfkknj', 'sfkkzy', 'sfznkx', 'zdkxms', 'kspage',
             'jspage']
    _params = {
        "rwlx": "1",
        "xkly": "1",
        "kklxdm": "01"
    }
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.args.get("su", None) != SU:
        return ""
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    for k, v in _params.items():
        if request.form.get(k, None) != v:
            return ""
    session["query_courses"] = session.get("query_courses", 0) + 1
    return app.send_static_file("zzxkyzb_cxZzxkYzbPartDisplay.html")


@app.route("/xsxk/zzxkyzb_cxJxbWithKchZzxkYzb.html", methods=["post"])
def selection_query_classes():
    _args = ['rwlx', 'xkly', 'xqh_id', 'zyh_id', 'njdm_id', 'bh_id', 'tykczgxdcs', 'xkxnm', 'xkxqm', 'kklxdm', 'xslbdm',
             'ccdm', 'kklxdm', 'xbm', 'zyfx_id', 'xsbj', 'kkbk', 'sfkknj', 'sfkkzy', 'sfznkx', 'zdkxms', 'kch_id']
    _params = {
        "rwlx": "1",
        "xkly": "1",
        "kklxdm": "01",
        "kch_id": "CS241"
    }
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    if request.args.get("su", None) != SU:
        return ""
    for arg in _args:
        if arg not in request.form.keys():
            return ""
    for k, v in _params.items():
        if request.form.get(k, None) != v:
            return ""
    session["query_classes"] = session.get("query_classes", 0) + 1
    return app.send_static_file("zzxkyzb_cxJxbWithKchZzxkYzb.html")


@app.route("/xsxk/zzxkyzb_xkBcZyZzxkYzb.html", methods=["post"])
def selection_register():
    _params = {
        "jxb_ids": "0f40b5296313cee8407cc4d78b0f8e17a22dd66977e4b2eba40bfa7e9609a2af5655bd56084525b52eaccc0022135fd40796bc9a0bb7246d827b52abe9404552947bb23a54841d74982f19b9a3cdc3c96455efeeb8bdeaef5a90b5afdb06f81b83de2008ab1be401d9f4d2b0a42b570de845e82f6d15a2e03c3e5ec42ebdcf3d",
        "kch_id": "CS241",
        "qz": "0"
    }
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for k, v in _params.items():
        if request.form.get(k, None) != v:
            return '{"msg":"出现未知异常，请与管理员联系！","flag":"0"}'
    if not session.get("no_conflict", False):
        return '{"msg": "所选教学班的上课时间与其他教学班有冲突！","flag":"0"}'
    if not session.get("no_full", False):
        return '{"msg":"1,A0000000000020C2E055F8163ED16360,80","flag":"-1"}'
    session["registered"] = True
    return '{"flag":"1"}'


@app.route("/xsxk/zzxkyzb_tuikBcZzxkYzb.html", methods=["post"])
def selection_deregister():
    _params = {
        "kch_id": "CS241",
        "jxb_ids": "0f40b5296313cee8407cc4d78b0f8e17a22dd66977e4b2eba40bfa7e9609a2af5655bd56084525b52eaccc0022135fd40796bc9a0bb7246d827b52abe9404552947bb23a54841d74982f19b9a3cdc3c96455efeeb8bdeaef5a90b5afdb06f81b83de2008ab1be401d9f4d2b0a42b570de845e82f6d15a2e03c3e5ec42ebdcf3d"
    }
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for k, v in _params.items():
        if request.form.get(k, None) != v:
            return '"3"'
    session["registered"] = False
    return '"1"'


@app.route("/xsxk/zzxkyzb_xkJcInXksjZzxkYzb.html", methods=["post"])
def selection_is_registered():
    _params = {
        "jxb_id": "0f40b5296313cee8407cc4d78b0f8e17a22dd66977e4b2eba40bfa7e9609a2af5655bd56084525b52eaccc0022135fd40796bc9a0bb7246d827b52abe9404552947bb23a54841d74982f19b9a3cdc3c96455efeeb8bdeaef5a90b5afdb06f81b83de2008ab1be401d9f4d2b0a42b570de845e82f6d15a2e03c3e5ec42ebdcf3d",
        "xkkz_id": "A000000000000B76E055F8163ED16360",
        "xnm": "2020",
        "xqm": "3"
    }
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    for k, v in _params.items():
        if request.form.get(k, None) != v:
            return '"0"'
    return '"1"' if session.get("registered", False) else '"0"'

@app.route("/xsxxxggl/xsgrxxwh_cxXsgrxx.html", methods=["get"])
def profile():
    if not session.get("is_login", None):
        return redirect("/xtgl/login_slogin.html")
    _args = {
        "gnmkdm": "N100801"
    }
    if request.args.get("su", None) != SU:
        return ""
    for k, v in _args.items():
        if request.args.get(k, None) != v:
            return ""
    session["query_profile"] = session.get("query_profile", 0) + 1
    return app.send_static_file("xsgrxxwh_cxXsgrxx.html")


import typing

from marshmallow import EXCLUDE, Schema, fields, post_load  # type: ignore

from pysjtu.schemas.base import ChineseBool


class ScoreFactorName(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return value[:value.find("(")]


class ScoreFactorPercentage(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return float(value[value.find("(") + 1:value.find("%")]) / 100


class ScoreFactorSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = ScoreFactorName(required=True, data_key="xmblmc")
    percentage = ScoreFactorPercentage(required=True, data_key="xmblmc", load_only=True)
    score = fields.String(required=True, data_key="xmcj")

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return ScoreFactor(**data)


class ScoreSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.Str(required=True, data_key="kcmc")
    teacher = fields.Str(data_key="jsxm")
    score = fields.Str(required=True, data_key="cj")
    credit = fields.Float(required=True, data_key="xf")
    gp = fields.Float(required=True, data_key="jd")
    invalid = ChineseBool(data_key="cjsfzf")
    course_type = fields.Str(data_key="kcbj")
    category = fields.Str(data_key="kclbmc")
    score_type = fields.Str(data_key="ksxz")
    method = fields.Str(data_key="khfsmc")
    course_id = fields.Str(data_key="kch_id")
    class_name = fields.Str(data_key="jxbmc")
    class_id = fields.Str(data_key="jxb_id")

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return Score(**data)


from pysjtu.models.score import Score, ScoreFactor



from pysjtu.ocr import LegacyRecognizer, NNRecognizer, JCSSRecognizer
from .client import Client, create_client
from .models import CourseRange, LogicEnum, Ranking
from .session import Session

__version__ = "0.3.4"
